@@ -8,7 +8,7 @@ must inherit from :class:`DesktopNotifierBase`.
 import logging
 from enum import Enum
 from collections import deque
-from typing import Optional, Dict, Callable, Union, Deque, List
+from typing import Optional, Dict, Callable, Any, Union, Deque, List
 
 
 logger = logging.getLogger(__name__)
@@ -35,16 +35,20 @@ class Notification:
 
     :param title: Notification title.
     :param message: Notification message.
-    :param urgency: Notification level: low, normal or critical. This is ignored by some
-        implementations.
+    :param urgency: Notification level: low, normal or critical.
     :param icon: Path to an icon to use for the notification, typically the app icon.
-        This is ignored by some implementations, e.g., on macOS where the icon of the
-        app bundle is always used.
-    :param action: Handler to call when the notification is clicked. This is ignored by
-        some implementations.
-    :param buttons: A dictionary with button names to show in the notification and
-        handler to call when the respective button is clicked. This is ignored by some
-        implementations.
+    :param buttons: A dictionary with button titles and callbacks to show in the
+        notification.
+    :param reply_field: Whether to show a reply field.
+    :param on_clicked: Callback to call when the notification is clicked. The
+        callback will be called without any arguments.
+    :param on_dismissed: Callback to call when the notification is dismissed. The
+        callback will be called without any arguments.
+    :param on_replied: If ``reply_field`` is True, a callback to call once the
+        user has replied. The callback will be called a single argument: a string
+        with the user reply.
+    :param sound: Whether to play a sound when the notification is shown.
+    :param thread: An identifier to group related notifications together.
     """
 
     def __init__(
@@ -53,9 +57,14 @@ class Notification:
         message: str,
         urgency: NotificationLevel = NotificationLevel.Normal,
         icon: Optional[str] = None,
-        action: Optional[Callable] = None,
-        buttons: Optional[Dict[str, Callable]] = None,
+        buttons: Optional[Dict[str, Callable[[], Any]]] = None,
+        reply_field: bool = False,
+        on_clicked: Optional[Callable[[], Any]] = None,
+        on_dismissed: Optional[Callable[[], Any]] = None,
+        on_replied: Optional[Callable[[str], Any]] = None,
+        attachment: Optional[str] = None,
         sound: bool = False,
+        thread: Optional[str] = None,
     ) -> None:
 
         self._identifier: Union[str, int, None] = None
@@ -63,9 +72,14 @@ class Notification:
         self.message = message
         self.urgency = urgency
         self.icon = icon
-        self.action = action
         self.buttons = buttons or dict()
+        self.reply_field = reply_field
+        self.on_clicked = on_clicked
+        self.on_dismissed = on_dismissed
+        self.on_replied = on_replied
+        self.attachment = attachment
         self.sound = sound
+        self.thread = thread
 
     @property
     def identifier(self) -> Union[str, int, None]:
@@ -105,6 +119,19 @@ class DesktopNotifierBase:
         self._current_notifications: Deque[Notification] = deque([], notification_limit)
         self._notification_for_nid: Dict[Union[str, int], Notification] = {}
 
+    def request_authorisation(self, callback: Optional[Callable]) -> None:
+        """
+        Request authorisation to send notifications.
+        """
+        raise NotImplementedError()
+
+    @property
+    def has_authorisation(self) -> bool:
+        """
+        Whether we have authorisation to send notifications.
+        """
+        raise NotImplementedError()
+
     def send(self, notification: Notification) -> None:
         """
         Sends a desktop notification. Some arguments may be ignored, depending on the
@@ -140,6 +167,22 @@ class DesktopNotifierBase:
             self._current_notifications.append(notification)
             self._notification_for_nid[platform_nid] = notification
 
+    def _clear_notification_from_cache(self, notification: Notification) -> None:
+        """
+        Removes the notification from our cache. Should be called by backends when the
+        notification is closed.
+        """
+        try:
+            self._current_notifications.remove(notification)
+        except ValueError:
+            pass
+
+        if notification.identifier:
+            try:
+                self._notification_for_nid.pop(notification.identifier)
+            except KeyError:
+                pass
+
     def _send(
         self,
         notification: Notification,
@@ -174,8 +217,8 @@ class DesktopNotifierBase:
 
         if notification.identifier:
             self._clear(notification)
-            self._current_notifications.remove(notification)
-            self._notification_for_nid.pop(notification.identifier)
+
+        self._clear_notification_from_cache(notification)
 
     def _clear(self, notification: Notification) -> None:
         """
