@@ -8,7 +8,7 @@ event loop.
 # system imports
 import asyncio
 import logging
-from typing import Optional, Coroutine, TypeVar
+from typing import Optional, Coroutine, TypeVar, Callable
 
 # external imports
 from dbus_next import Variant  # type: ignore
@@ -24,6 +24,11 @@ logger = logging.getLogger(__name__)
 
 T = TypeVar("T")
 
+NOTIFICATION_CLOSED_EXPIRED = 1
+NOTIFICATION_CLOSED_DISMISSED = 2
+NOTIFICATION_CLOSED_PROGRAMMATICALLY = 3
+NOTIFICATION_CLOSED_UNDEFINED = 4
+
 
 class DBusDesktopNotifier(DesktopNotifierBase):
     """DBus notification backend for Linux
@@ -56,6 +61,19 @@ class DBusDesktopNotifier(DesktopNotifierBase):
         self._loop = asyncio.get_event_loop()
         self.interface: Optional[ProxyInterface] = None
 
+    def request_authorisation(self, callback: Optional[Callable]) -> None:
+        """
+        Request authorisation to send notifications.
+        """
+        pass
+
+    @property
+    def has_authorisation(self) -> bool:
+        """
+        Whether we have authorisation to send notifications.
+        """
+        return True
+
     def _run_coco_sync(self, coro: Coroutine[None, None, T]) -> T:
         """
         Runs the given coroutine and returns the result synchronously.
@@ -83,6 +101,9 @@ class DBusDesktopNotifier(DesktopNotifierBase):
         self.interface = self.proxy_object.get_interface(
             "org.freedesktop.Notifications"
         )
+
+        self.interface.on_notification_closed(self._on_closed)
+
         if hasattr(self.interface, "on_action_invoked"):
             # some older interfaces may not support notification actions
             self.interface.on_action_invoked(self._on_action)
@@ -129,11 +150,15 @@ class DBusDesktopNotifier(DesktopNotifierBase):
         for button_name in notification.buttons.keys():
             actions += [button_name, button_name]
 
+        hints = {"urgency": self._to_native_urgency[notification.urgency]}
+
         # sound
         if notification.sound:
-            sound = Variant("s", "message-new-instant")
-        else:
-            sound = Variant("s", "")
+            hints["sound-name"] = Variant("s", "message-new-instant")
+
+        # attachment
+        if notification.attachment:
+            hints["image-path"] = Variant("s", notification.attachment)
 
         # Post the new notification and record the platform ID assigned to it.
         platform_nid = await self.interface.call_notify(
@@ -143,32 +168,12 @@ class DBusDesktopNotifier(DesktopNotifierBase):
             notification.title,  # summary
             notification.message,  # body
             actions,  # actions
-            {
-                "urgency": self._to_native_urgency[notification.urgency],
-                "sound-name": sound,
-            },  # hints
+            hints,  # hints
             -1,  # expire_timeout (-1 = default)
         )
 
         return platform_nid
 
-    def _on_action(self, nid, action_key) -> None:
-
-        # Get the notification instance from the platform ID.
-        nid = int(nid)
-        action_key = str(action_key)
-        notification = self._notification_for_nid[nid]
-
-        # Execute any callbacks for button clicks.
-        if notification:
-            if action_key == "default" and notification.action:
-                notification.action()
-            else:
-                callback = notification.buttons.get(action_key)
-
-                if callback:
-                    callback()
-
     def _clear(self, notification: Notification) -> None:
         """
         Synchronously removes a notifications from the notification center
@@ -209,3 +214,41 @@ class DBusDesktopNotifier(DesktopNotifierBase):
 
         for notification in self.current_notifications:
             await self.interface.call_close_notification(notification.identifier)
+
+    # Note that _on_action and _on_closed might be called for the same notification
+    # with some notification servers. This is not a problem because the _on_action
+    # call will come first, in which case we are no longer interested in calling the
+    # on_dismissed callback.
+
+    def _on_action(self, nid, action_key) -> None:
+
+        # Get the notification instance from the platform ID.
+        nid = int(nid)
+        action_key = str(action_key)
+        notification = self._notification_for_nid.get(nid)
+
+        # Execute any callbacks for button clicks.
+        if notification:
+            self._clear_notification_from_cache(notification)
+
+            if action_key == "default" and notification.on_clicked:
+                notification.on_clicked()
+            else:
+                callback = notification.buttons.get(action_key)
+
+                if callback:
+                    callback()
+
+    def _on_closed(self, nid, reason) -> None:
+
+        # Get the notification instance from the platform ID.
+        nid = int(nid)
+        reason = int(reason)
+        notification = self._notification_for_nid.get(nid)
+
+        # Execute callback for user dismissal.
+        if notification:
+            self._clear_notification_from_cache(notification)
+
+            if reason == NOTIFICATION_CLOSED_DISMISSED and notification.on_dismissed:
+                notification.on_dismissed()
