@@ -5,7 +5,7 @@ Dummy backend for unsupported platforms.
 
 # system imports
 import uuid
-from typing import Optional
+from typing import Optional, Callable
 
 # local imports
 from .base import Notification, DesktopNotifierBase
@@ -22,6 +22,19 @@ class DummyNotificationCenter(DesktopNotifierBase):
     ) -> None:
         super().__init__(app_name, app_icon, notification_limit)
 
+    def request_authorisation(self, callback: Optional[Callable]) -> None:
+        """
+        Request authorisation to send notifications.
+        """
+        pass
+
+    @property
+    def has_authorisation(self) -> bool:
+        """
+        Whether we have authorisation to send notifications.
+        """
+        return True
+
     def _send(
         self,
         notification: Notification,
