@@ -13,8 +13,8 @@ UNUserNotificationCenter backend for macOS.
 # system imports
 import uuid
 import logging
-from concurrent.futures import Future, wait
-from typing import Optional, Dict, Tuple
+from concurrent.futures import Future
+from typing import Optional, Callable, Any, cast
 
 # external imports
 from rubicon.objc import NSObject, ObjCClass, objc_method, py_from_ns  # type: ignore
@@ -35,9 +35,12 @@ UNUserNotificationCenter = ObjCClass("UNUserNotificationCenter")
 UNMutableNotificationContent = ObjCClass("UNMutableNotificationContent")
 UNNotificationRequest = ObjCClass("UNNotificationRequest")
 UNNotificationAction = ObjCClass("UNNotificationAction")
+UNTextInputNotificationAction = ObjCClass("UNTextInputNotificationAction")
 UNNotificationCategory = ObjCClass("UNNotificationCategory")
 UNNotificationSound = ObjCClass("UNNotificationSound")
+UNNotificationAttachment = ObjCClass("UNNotificationAttachment")
 
+NSURL = ObjCClass("NSURL")
 NSSet = ObjCClass("NSSet")
 
 UNNotificationDefaultActionIdentifier = (
@@ -51,7 +54,10 @@ UNAuthorizationOptionBadge = 1 << 0
 UNAuthorizationOptionSound = 1 << 1
 UNAuthorizationOptionAlert = 1 << 2
 
+UNNotificationActionOptionAuthenticationRequired = 1 << 0
+UNNotificationActionOptionDestructive = 1 << 1
 UNNotificationActionOptionForeground = 1 << 2
+UNNotificationActionOptionNone = 0
 
 UNNotificationCategoryOptionNone = 0
 
@@ -60,6 +66,9 @@ UNAuthorizationStatusProvisional = 3
 UNAuthorizationStatusEphemeral = 4
 
 
+ReplyActionIdentifier = "com.desktop-notifier.ReplyActionIdentifier"
+
+
 class NotificationCenterDelegate(NSObject):  # type: ignore
     """Delegate to handle user interactions with notifications"""
 
@@ -71,19 +80,30 @@ class NotificationCenterDelegate(NSObject):  # type: ignore
         # Get the notification which was clicked from the platform ID.
         platform_nid = py_from_ns(response.notification.request.identifier)
         py_notification = self.interface._notification_for_nid[platform_nid]
+        py_notification = cast(Notification, py_notification)
 
-        # Get and call the callback which corresponds to the user interaction.
+        self.interface._clear_notification_from_cache(py_notification)
+
+        # Invoke the callback which corresponds to the user interaction.
         if response.actionIdentifier == UNNotificationDefaultActionIdentifier:
 
-            callback = py_notification.action
+            if py_notification.on_clicked:
+                py_notification.on_clicked()
 
-            if callback:
-                callback()
+        elif response.actionIdentifier == UNNotificationDismissActionIdentifier:
 
-        elif response.actionIdentifier != UNNotificationDismissActionIdentifier:
+            if py_notification.on_dismissed:
+                py_notification.on_dismissed()
+
+        elif response.actionIdentifier == ReplyActionIdentifier:
+
+            if py_notification.on_replied:
+                reply_text = py_from_ns(response.userText)
+                py_notification.on_replied(reply_text)
+
+        else:
 
             action_id_str = py_from_ns(response.actionIdentifier)
-
             callback = py_notification.buttons.get(action_id_str)
 
             if callback:
@@ -119,25 +139,29 @@ class CocoaNotificationCenter(DesktopNotifierBase):
         self.nc_delegate.interface = self
         self.nc.delegate = self.nc_delegate
 
-        self._did_request_authorisation = False
-        self._notification_categories: Dict[Tuple[str, ...], str] = {}
+        self._clear_notification_categories()
 
-    def request_authorisation(self) -> None:
+    def request_authorisation(
+        self, callback: Optional[Callable[[bool, str], Any]] = None
+    ) -> None:
         """
         Request authorisation to send user notifications. This method returns
         immediately but authorisation will only be granted once the user has accepted
-        the prompt. Use :attr:`has_authorisation` to check if we are authorised.
+        the prompt. Use :attr:`has_authorisation` to check if we are authorised or pass
+        a callback to be called when the request has been processed.
+
+        :param callback: A method to call when the authorisation request has been
+            granted or denied. The callback will be called with two arguments: a bool
+            indicating if authorisation was granted and a string describing failure
+            reasons for the request.
         """
 
         def on_auth_completed(granted: bool, error: objc_id) -> None:
-            if granted:
-                logger.debug("Authorisation granted")
-            else:
-                logger.debug("Authorisation denied")
 
-            if error:
-                error = py_from_ns(error)
-                logger.warning("Authorisation error %s", str(error))
+            if callback:
+                ns_error = py_from_ns(error)
+                error_description = ns_error.localizedDescription if ns_error else ""
+                callback(granted, error_description)
 
         self.nc.requestAuthorizationWithOptions(
             UNAuthorizationOptionAlert
@@ -146,13 +170,11 @@ class CocoaNotificationCenter(DesktopNotifierBase):
             completionHandler=on_auth_completed,
         )
 
-        self._did_request_authorisation = True
-
     @property
     def has_authorisation(self) -> bool:
         """Whether we have authorisation to send notifications."""
 
-        # get existing notification categories
+        # Get existing notification categories.
 
         future: Future = Future()
 
@@ -163,7 +185,6 @@ class CocoaNotificationCenter(DesktopNotifierBase):
 
         self.nc.getNotificationSettingsWithCompletionHandler(handler)
 
-        wait([future])
         settings = future.result()
 
         authorized = settings.authorizationStatus in (
@@ -188,9 +209,6 @@ class CocoaNotificationCenter(DesktopNotifierBase):
         :param notification_to_replace: Notification to replace, if any.
         """
 
-        if not self._did_request_authorisation:
-            self.request_authorisation()
-
         if not self.has_authorisation:
             raise RuntimeError("Not authorised")
 
@@ -199,72 +217,99 @@ class CocoaNotificationCenter(DesktopNotifierBase):
         else:
             platform_nid = str(uuid.uuid4())
 
-        # Set up buttons for notification. On macOS, we need need to register a new
-        # notification category for every unique set of buttons.
-        button_names = tuple(notification.buttons.keys())
-        category_id = self._category_id_for_button_names(button_names)
+        # On macOS, we need need to register a new notification category for every
+        # unique set of buttons.
+        category_id = self._create_category_for_notification(notification)
 
         # Create the native notification + notification request.
         content = UNMutableNotificationContent.alloc().init()
         content.title = notification.title
         content.body = notification.message
         content.categoryIdentifier = category_id
+        content.threadIdentifier = notification.thread
 
         if notification.sound:
             content.sound = UNNotificationSound.defaultSound
 
+        if notification.attachment:
+            url = NSURL.fileURLWithPath(notification.attachment)
+            attachment = UNNotificationAttachment.attachmentWithIdentifier(
+                "", URL=url, options={}, error=None
+            )
+            content.attachments = [attachment]
+
         notification_request = UNNotificationRequest.requestWithIdentifier(
             platform_nid, content=content, trigger=None
         )
 
+        future: Future = Future()
+
+        def handler(error: objc_id) -> None:
+            ns_error = py_from_ns(error)
+            error_description = ns_error.localizedDescription if ns_error else ""
+            future.set_result(error_description)
+
         # Post the notification.
-        self.nc.addNotificationRequest(notification_request, withCompletionHandler=None)
+        self.nc.addNotificationRequest(
+            notification_request, withCompletionHandler=handler
+        )
+
+        error = future.result()
+
+        if error != "":
+            raise RuntimeError(error)
 
         return platform_nid
 
-    def _category_id_for_button_names(
-        self, button_names: Tuple[str, ...]
+    def _create_category_for_notification(
+        self, notification: Notification
     ) -> Optional[str]:
         """
-        Creates and registers a new notification category with the given buttons
-        or retrieves an existing one.
+        Registers a new notification category with UNNotificationCenter for the given
+        notification or retrieves an existing one if it exists for our set of buttons.
 
-        :returns: UNUserNotificationCenter ID for a new or an exiting category. Returns
-            None if a notification category without buttons is requested.
+        :param notification: Notification instance.
+        :returns: The identifier of the existing or created notification category.
         """
 
-        if not button_names:
+        if not (notification.buttons or notification.reply_field):
             return None
 
-        try:
-            return self._notification_categories[button_names]
-        except KeyError:
+        button_titles = tuple(notification.buttons)
+        ui_repr = f"buttons={button_titles}, reply_field={notification.reply_field}"
+        category_id = f"desktop-notifier: {ui_repr}"
+
+        # Retrieve existing categories. We do not cache this value because it may be
+        # modified by other Python processes using desktop-notifier.
+
+        categories = self._get_notification_categories()
+        category_ids = set(py_from_ns(c.identifier) for c in categories.allObjects())  # type: ignore
+
+        # Register new category if necessary.
+        if category_id not in category_ids:
+
+            # Create action for each button.
             actions = []
 
-            for name in button_names:
-                action = UNNotificationAction.actionWithIdentifier(
-                    name, title=name, options=UNNotificationActionOptionForeground
+            if notification.reply_field:
+                action = UNTextInputNotificationAction.actionWithIdentifier(
+                    ReplyActionIdentifier,
+                    title="Reply",
+                    options=UNNotificationActionOptionNone,
+                    textInputButtonTitle="Reply",
+                    textInputPlaceholder="",
                 )
                 actions.append(action)
 
-            # get existing notification categories
+            for name in notification.buttons:
+                action = UNNotificationAction.actionWithIdentifier(
+                    name, title=name, options=UNNotificationActionOptionNone
+                )
+                actions.append(action)
 
-            future: Future = Future()
+            # Add category for new set of buttons.
 
-            def handler(categories: objc_id) -> None:
-                categories = py_from_ns(categories)
-                categories.retain()
-                future.set_result(categories)
-
-            self.nc.getNotificationCategoriesWithCompletionHandler(handler)
-
-            wait([future])
-            categories = future.result()
-
-            # add category for new set of buttons
-
-            category_id = str(uuid.uuid4())
-            new_categories = categories.setByAddingObject(
+            new_categories = categories.setByAddingObject(  # type: ignore
                 UNNotificationCategory.categoryWithIdentifier(
                     category_id,
                     actions=actions,
@@ -273,11 +318,30 @@ class CocoaNotificationCenter(DesktopNotifierBase):
                 )
             )
             self.nc.setNotificationCategories(new_categories)
-            self._notification_categories[button_names] = category_id
 
-            categories.release()
+        return category_id
 
-            return category_id
+    def _get_notification_categories(self) -> NSSet:  # type: ignore
+        """Returns the registered notification categories for this app / Python."""
+
+        future: Future = Future()
+
+        def handler(categories: objc_id) -> None:
+            categories = py_from_ns(categories)
+            categories.retain()
+            future.set_result(categories)
+
+        self.nc.getNotificationCategoriesWithCompletionHandler(handler)
+
+        categories = future.result()
+        categories.autorelease()
+
+        return categories
+
+    def _clear_notification_categories(self) -> None:
+        """Clears all registered notification categories for this application."""
+        empty_set = NSSet.alloc().init()
+        self.nc.setNotificationCategories(empty_set)
 
     def _clear(self, notification: Notification) -> None:
         """
