@@ -12,7 +12,7 @@ NSUserNotificationCenter backend for macOS.
 import uuid
 import platform
 import logging
-from typing import Optional
+from typing import Optional, Callable, cast
 
 # external imports
 from rubicon.objc import NSObject, ObjCClass, objc_method, py_from_ns  # type: ignore
@@ -50,14 +50,17 @@ class NotificationCenterDelegate(NSObject):  # type: ignore
 
         platform_nid = py_from_ns(notification.identifier)
         py_notification = self.interface._notification_for_nid[platform_nid]
+        py_notification = cast(Notification, py_notification)
+
+        self.interface._clear_notification_from_cache(py_notification)
 
         if (
             notification.activationType
             == NSUserNotificationActivationTypeContentsClicked
         ):
 
-            if py_notification.action:
-                py_notification.action()
+            if py_notification.on_clicked:
+                py_notification.on_clicked()
 
         elif (
             notification.activationType
@@ -100,6 +103,19 @@ class CocoaNotificationCenterLegacy(DesktopNotifierBase):
         self.nc_delegate.interface = self
         self.nc.delegate = self.nc_delegate
 
+    def request_authorisation(self, callback: Optional[Callable]) -> None:
+        """
+        Request authorisation to send notifications.
+        """
+        pass
+
+    @property
+    def has_authorisation(self) -> bool:
+        """
+        Whether we have authorisation to send notifications.
+        """
+        return True
+
     def _send(
         self,
         notification: Notification,
